from typing import Any, Callable
from uuid import UUID

from fastapi import APIRouter, FastAPI
from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field, field_validator

from gen_epix.casedb.api.base import EXCLUDED_PERMISSIONS
from gen_epix.casedb.domain import command, enum, model
from gen_epix.fastapp import App
from gen_epix.fastapp.api import CrudEndpointGenerator


class UpdateCaseTypeSetCaseTypesRequestBody(PydanticBaseModel):
    case_type_set_members: list[model.CaseTypeSetMember]


class UpdateCaseTypeColSetCaseTypeColsRequestBody(PydanticBaseModel):
    case_type_col_set_members: list[model.CaseTypeColSetMember]


class CreateCasesRequestBody(PydanticBaseModel):
    cases: list[model.Case]
    data_collection_ids: set[UUID] = Field(
        description="The data collections in which the cases will be put initially. Must contain exactly one source data collection, plus any other data collections in which the cases may be shared from there.",
        min_length=1,
    )

    @field_validator("cases", mode="after")
    def _validate_cases(cls, value: list[model.Case]) -> list[model.Case]:
        if len(set(x.case_type_id for x in value)) > 1:
            raise ValueError("Not all cases have the same case type.")
        return value


class CreateCaseSetRequestBody(PydanticBaseModel):
    case_set: model.CaseSet
    data_collection_ids: set[UUID] = Field(
        description="The data collections in which the case set will be put initially. Must contain exactly one source data collection, plus any other data collections in which the case set may be shared from there.",
        min_length=1,
    )
    case_ids: set[UUID] | None = Field(
        default=None, description="The cases to be added to the case set, if any."
    )


class RetrieveOrganizationContactRequestBody(PydanticBaseModel):
    organization_ids: list[UUID] | None = None
    site_ids: list[UUID] | None = None
    contact_ids: list[UUID] | None = None
    props: dict[str, Any] = {}


class RetrievePhylogeneticTreeRequestBody(PydanticBaseModel):
    genetic_distance_case_type_col_id: UUID
    tree_algorithm_code: enum.TreeAlgorithmType
    case_ids: list[UUID]


class RetrieveGeneticSequenceRequestBody(PydanticBaseModel):
    genetic_sequence_case_type_col_id: UUID
    case_ids: list[UUID]


class RetrieveAlleleProfileRequestBody(PydanticBaseModel):
    sequence_ids: list[UUID]
    props: dict[str, Any] = {}


def create_case_endpoints(
    router: APIRouter | FastAPI,
    app: App,
    registered_user_dependency: Callable | None = None,
    new_user_dependency: Callable | None = None,
    idp_user_dependency: Callable | None = None,
    handle_exception: Callable | None = None,
    **kwargs: dict,
) -> None:
    assert handle_exception

    # Specific endpoints - Case
    @router.put(
        "/case_type_sets/{case_type_set_id}/case_types",
        operation_id="case_type_sets__put__case_types",
        name="CaseTypeSet_CaseType",
    )
    async def case_type_sets__put__case_types(
        user: registered_user_dependency,  # type: ignore
        case_type_set_id: UUID,
        request_body: UpdateCaseTypeSetCaseTypesRequestBody,
    ) -> list[model.CaseSetMember]:
        try:
            cmd = command.CaseTypeSetCaseTypeUpdateAssociationCommand(
                user=user,
                obj_id1=case_type_set_id,
                association_objs=request_body.case_type_set_members,
                props={"return_id": False},
            )
            retval: list[model.CaseSetMember] = app.handle(cmd)
        except Exception as exception:
            handle_exception("fbe272b9", user, exception)
        return retval

    @router.put(
        "/case_type_col_sets/{case_type_col_set_id}/case_type_cols",
        operation_id="case_type_col_sets__put__case_type_cols",
        name="CaseTypeColSet_CaseTypeCol",
    )
    async def case_type_col_sets__put__case_type_cols(
        user: registered_user_dependency,  # type: ignore
        case_type_col_set_id: UUID,
        request_body: UpdateCaseTypeColSetCaseTypeColsRequestBody,
    ) -> list[model.CaseTypeColSetMember]:
        try:
            cmd = command.CaseTypeColSetCaseTypeColUpdateAssociationCommand(
                user=user,
                obj_id1=case_type_col_set_id,
                association_objs=request_body.case_type_col_set_members,
                props={"return_id": False},
            )
            retval: list[model.CaseTypeColSetMember] = app.handle(cmd)
        except Exception as exception:
            handle_exception("ab010768", user, exception)
        return retval

    @router.get(
        "/complete_case_types",
        operation_id="complete_case_types__get_one",
        name="CompleteCaseType",
    )
    async def complete_case_types__get_one(
        user: registered_user_dependency,  # type: ignore
        case_type_id: UUID,
    ) -> model.CompleteCaseType:
        try:
            cmd = command.RetrieveCompleteCaseTypeCommand(
                user=user, case_type_id=case_type_id
            )
            retval: model.CompleteCaseType = app.handle(cmd)
        except Exception as exception:
            handle_exception("c6c17125", user, exception)
        return retval

    @router.post(
        "/create/cases",
        operation_id="create__cases",
        name="Cases",
    )
    async def create__cases(
        user: registered_user_dependency,  # type: ignore
        request_body: CreateCasesRequestBody,
    ) -> list[model.Case]:
        try:
            cmd = command.CasesCreateCommand(
                user=user,
                cases=request_body.cases,
                data_collection_ids=request_body.data_collection_ids,
            )
            retval: list[model.Case] = app.handle(cmd)
        except Exception as exception:
            handle_exception("b413ab76", user, exception)
        return retval

    @router.post(
        "/create/case_set",
        operation_id="create__case_set",
        name="CaseSet",
    )
    async def create__case_set(
        user: registered_user_dependency,  # type: ignore
        request_body: CreateCaseSetRequestBody,
    ) -> model.CaseSet:
        try:
            cmd = command.CaseSetCreateCommand(
                user=user,
                case_set=request_body.case_set,
                data_collection_ids=request_body.data_collection_ids,
                case_ids=request_body.case_ids,
            )
            retval: model.CaseSet = app.handle(cmd)
        except Exception as exception:
            handle_exception("c39c42f9", user, exception)
        return retval

    @router.post(
        "/retrieve/case_type_stats",
        operation_id="retrieve__case_type_stats",
        name="CaseTypeStats",
    )
    async def retrieve__case_type_stats(
        user: registered_user_dependency,  # type: ignore
        request_body: command.RetrieveCaseTypeStatsCommand,
    ) -> list[model.CaseTypeStat]:
        try:
            cmd = request_body
            cmd.user = user
            retval: list[model.CaseTypeStat] = app.handle(cmd)
        except Exception as exception:
            handle_exception("80c99f53", user, exception)
        return retval

    @router.get(
        "/retrieve/case_set_stats",
        operation_id="retrieve__case_set_stats",
        name="CaseSetStats",
    )
    async def retrieve__case_set_stats(
        user: registered_user_dependency,  # type: ignore
    ) -> list[model.CaseSetStat]:
        try:
            cmd = command.RetrieveCaseSetStatsCommand(user=user)
            retval: list[model.CaseSetStat] = app.handle(cmd)
        except Exception as exception:
            handle_exception("be54843e", user, exception)
        return retval

    @router.post(
        "/retrieve/case_ids_by_query", operation_id="retrieve__case_ids_by_query"
    )
    async def retrieve__case_ids_by_query(
        user: registered_user_dependency,  # type: ignore
        request_body: model.CaseQuery,
    ) -> list[UUID]:
        try:
            retval: list[UUID] = app.handle(
                command.RetrieveCasesByQueryCommand(
                    user=user,
                    case_query=request_body,
                )
            )
        except Exception as exception:
            handle_exception("a8f773fe", user, exception)
        return retval

    @router.post("/retrieve/cases_by_ids", operation_id="retrieve__cases_by_ids")
    async def retrieve__cases_by_ids(
        user: registered_user_dependency,  # type: ignore
        request_body: list[UUID],
    ) -> list[model.Case]:
        try:
            retval: list[model.Case] = app.handle(
                command.RetrieveCasesByIdCommand(
                    user=user,
                    case_ids=request_body,
                )
            )
        except Exception as exception:
            handle_exception("f6d423fe", user, exception)
        return retval

    @router.post(
        "/retrieve/case_rights",
        operation_id="retrieve__case_rights",
    )
    async def retrieve__case_rights(
        user: registered_user_dependency,  # type: ignore
        request_body: list[UUID],
    ) -> list[model.CaseRights]:
        try:
            retval: list[model.CaseRights] = app.handle(
                command.RetrieveCaseRightsCommand(
                    user=user,
                    case_ids=request_body,
                )
            )
        except Exception as exception:
            handle_exception("c6f4b3c2", user, exception)
        return retval

    @router.post(
        "/retrieve/case_set_rights",
        operation_id="retrieve__case_set_rights",
    )
    async def retrieve__case_set_rights(
        user: registered_user_dependency,  # type: ignore
        request_body: list[UUID],
    ) -> list[model.CaseSetRights]:
        try:
            retval: list[model.CaseSetRights] = app.handle(
                command.RetrieveCaseSetRightsCommand(
                    user=user,
                    case_set_ids=request_body,
                )
            )
        except Exception as exception:
            handle_exception("b9c49fe1", user, exception)
        return retval

    @router.post(
        "/retrieve/case_set_rights",
        operation_id="retrieve__case_set_rights",
    )
    async def retrieve__case_set_rights(
        user: registered_user_dependency,  # type: ignore
        request_body: list[UUID],
    ) -> list[model.CaseSetRights]:
        try:
            retval: list[model.CaseSetRights] = app.handle(
                command.RetrieveCaseSetRightsCommand(
                    user=user,
                    case_set_ids=request_body,
                )
            )
        except Exception as e:
            handle_exception("b9c49fe1", user, e)
        return retval

    @router.post(
        "/retrieve/organization_contact", operation_id="retrieve__organization_contact"
    )
    async def retrieve__organization_contact(
        user: registered_user_dependency, request_body: RetrieveOrganizationContactRequestBody  # type: ignore
    ) -> list[model.Contact]:
        try:
            retval: list[model.Contact] = app.handle(
                command.RetrieveOrganizationContactCommand(
                    user=user,
                    organization_ids=request_body.organization_ids,
                    site_ids=request_body.site_ids,
                    contact_ids=request_body.contact_ids,
                    props=request_body.props,
                )
            )
        except Exception as exception:
            handle_exception(
                "b8172f62",
                user,
                exception,
                request_ids=(request_body.organization_ids or [])
                + (request_body.site_ids or [])
                + (request_body.contact_ids or []),
            )
        return retval

    @router.post(
        "/retrieve/phylogenetic_tree", operation_id="retrieve__phylogenetic_tree"
    )
    async def retrieve__phylogenetic_tree(
        user: registered_user_dependency, request_body: RetrievePhylogeneticTreeRequestBody  # type: ignore
    ) -> model.PhylogeneticTree:
        try:
            retval: model.PhylogeneticTree = app.handle(
                command.RetrievePhylogeneticTreeByCasesCommand(
                    user=user,
                    genetic_distance_case_type_col_id=request_body.genetic_distance_case_type_col_id,
                    tree_algorithm=request_body.tree_algorithm_code,
                    case_ids=request_body.case_ids,
                )
            )
        except Exception as exception:
            handle_exception(
                "45219a88", user, exception, request_ids=request_body.case_ids
            )
        return retval

    @router.post(
        "/retrieve/genetic_sequence", operation_id="retrieve__genetic_sequence"
    )
    async def retrieve__genetic_sequence(
        user: registered_user_dependency,  # type: ignore
        request_body: RetrieveGeneticSequenceRequestBody,
    ) -> list[model.GeneticSequence]:
        try:
            retval: list[model.GeneticSequence] = app.handle(
                command.RetrieveGeneticSequenceByCaseCommand(
                    user=user,
                    genetic_sequence_case_type_col_id=request_body.genetic_sequence_case_type_col_id,
                    case_ids=request_body.case_ids,
                )
            )
        except Exception as exception:
            handle_exception(
                "1238afb2", user, exception, request_ids=request_body.case_ids
            )
        return retval

    # @router.post("/retrieve/allele_profile", operation_id="retrieve__allele_profile")
    # async def retrieve__allele_profile(
    #     user: registered_user_dependency, request_body: RetrieveAlleleProfileRequestBody  # type: ignore
    # ) -> list[model.SeqDbAlleleProfile]:
    #     try:
    #         retval: list[model.SeqDbAlleleProfile] = app.handle(
    #             command.RetrieveAlleleProfileCommand(
    #                 user=user,
    #                 sequence_ids=request_body.sequence_ids,
    #                 props=request_body.props,
    #             )
    #         )
    #     except Exception as exception:
    #         handle_exception("a4c03b54", user, exception, request_ids=request_body.sequence_ids)
    #     return retval

    # CRUD
    crud_endpoint_sets = CrudEndpointGenerator.create_crud_endpoint_set_for_domain(
        app,
        service_type=enum.ServiceType.CASE,
        user_dependency=registered_user_dependency,
        excluded_permissions=EXCLUDED_PERMISSIONS,
    )
    CrudEndpointGenerator.generate_endpoints(
        router, crud_endpoint_sets, handle_exception
    )



from uuid import UUID

from pydantic import BaseModel

import gen_epix.casedb.domain.model as model
from gen_epix.casedb.domain.model import Model
from gen_epix.fastapp import PermissionTypeSet

EXCLUDED_PERMISSIONS: dict = {
    model.User: PermissionTypeSet.CU,
    model.UserInvitation: PermissionTypeSet.CU,
    model.CaseSet: PermissionTypeSet.C,
}


class UpdateAssociationRequestBody(BaseModel):
    obj_ids1: list[UUID] | UUID | None = None
    obj_ids2: list[UUID] | UUID | None = None
    association_objs: list[Model] | None = None



# pylint: disable=too-few-public-methods
# This module defines base classes, methods are added later


from typing import Any, Type
from uuid import UUID

import sqlalchemy as sa
from sqlalchemy.orm import Mapped, relationship

from gen_epix.casedb.domain import enum, model
from gen_epix.casedb.repositories.sa_model.base import RowMetadataMixin
from gen_epix.casedb.repositories.sa_model.util import (
    create_mapped_column,
    create_table_args,
)

Base: Type = sa.orm.declarative_base(name=enum.ServiceType.ONTOLOGY.value)


class Concept(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Concept)

    abbreviation: Mapped[str] = create_mapped_column(model.Concept, "abbreviation")
    name: Mapped[str] = create_mapped_column(model.Concept, "name")
    description: Mapped[str] = create_mapped_column(model.Concept, "description")
    props: Mapped[dict[str, Any]] = create_mapped_column(model.Concept, "props")


class ConceptSet(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.ConceptSet)

    code: Mapped[str] = create_mapped_column(model.ConceptSet, "code")
    name: Mapped[str] = create_mapped_column(model.ConceptSet, "name")
    type: Mapped[enum.ConceptSetType] = create_mapped_column(model.ConceptSet, "type")
    regex: Mapped[str] = create_mapped_column(model.ConceptSet, "regex")
    schema_definition: Mapped[str] = create_mapped_column(
        model.ConceptSet, "schema_definition"
    )
    schema_uri: Mapped[str] = create_mapped_column(model.ConceptSet, "schema_uri")
    description: Mapped[str] = create_mapped_column(model.ConceptSet, "description")


class ConceptSetMember(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.ConceptSetMember)

    concept_set_id: Mapped[UUID] = create_mapped_column(
        model.ConceptSetMember, "concept_set_id"
    )
    concept_id: Mapped[UUID] = create_mapped_column(
        model.ConceptSetMember, "concept_id"
    )
    rank: Mapped[int] = create_mapped_column(model.ConceptSetMember, "rank")

    concept_set: Mapped[ConceptSet] = relationship(
        ConceptSet, foreign_keys=[concept_set_id]
    )
    concept: Mapped[Concept] = relationship(Concept, foreign_keys=[concept_id])


class Disease(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Disease)

    name: Mapped[str] = create_mapped_column(model.Disease, "name")
    icd_code: Mapped[str] = create_mapped_column(model.Disease, "icd_code")


class EtiologicalAgent(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.EtiologicalAgent)

    name: Mapped[str] = create_mapped_column(model.EtiologicalAgent, "name")
    type: Mapped[str] = create_mapped_column(model.EtiologicalAgent, "type")


class Etiology(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Etiology)

    disease_id: Mapped[UUID] = create_mapped_column(model.Etiology, "disease_id")
    etiological_agent_id: Mapped[UUID] = create_mapped_column(
        model.Etiology, "etiological_agent_id"
    )

    disease: Mapped[Disease] = relationship(Disease, foreign_keys=[disease_id])
    etiological_agent: Mapped[EtiologicalAgent] = relationship(
        EtiologicalAgent, foreign_keys=[etiological_agent_id]
    )



# pylint: disable=useless-import-alias
from typing import Type

from gen_epix.fastapp.service import BaseService
from gen_epix.fastapp.services.auth.base import BaseAuthService
from gen_epix.omopdb.domain import enum
from gen_epix.omopdb.domain.service.omop import BaseOmopService as BaseOmopService
from gen_epix.omopdb.domain.service.organization import (
    BaseOrganizationService as BaseOrganizationService,
)
from gen_epix.omopdb.domain.service.rbac import BaseRbacService as BaseRbacService
from gen_epix.omopdb.domain.service.system import BaseSystemService as BaseSystemService

ORDERED_SERVICE_TYPES: list[enum.ServiceType] = [
    enum.ServiceType.ORGANIZATION,
    enum.ServiceType.AUTH,
    enum.ServiceType.SYSTEM,
    enum.ServiceType.OMOP,
    enum.ServiceType.RBAC,
]

BASE_SERVICE_CLASS_MAP: dict[enum.ServiceType, Type[BaseService]] = {
    enum.ServiceType.ORGANIZATION: BaseOrganizationService,
    enum.ServiceType.AUTH: BaseAuthService,
    enum.ServiceType.SYSTEM: BaseSystemService,
    enum.ServiceType.OMOP: BaseOmopService,
    enum.ServiceType.RBAC: BaseRbacService,
}



from gen_epix.fastapp import exc
from gen_epix.fastapp.repositories import DictRepository
from gen_epix.fastapp.unit_of_work import BaseUnitOfWork
from gen_epix.seqdb.domain import model
from gen_epix.seqdb.domain.repository.organization import BaseOrganizationRepository


class OrganizationDictRepository(DictRepository, BaseOrganizationRepository):
    def is_existing_user_by_key(
        self, uow: BaseUnitOfWork, user_key: str | None
    ) -> bool:
        if user_key is None:
            return False
        for user in self._db[model.User].values():
            assert isinstance(user, model.User)
            if user.email == user_key:
                return True
        return False

    def retrieve_user_by_key(self, uow: BaseUnitOfWork, user_key: str) -> model.User:
        for user in self._db[model.User].values():
            assert isinstance(user, model.User)
            if user.email == user_key.lower():
                return user
        raise exc.NoResultsError()



from typing import Callable
from uuid import UUID

from fastapi import APIRouter, FastAPI
from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field

from gen_epix.fastapp import App
from gen_epix.fastapp.api.crud_endpoint_generator import CrudEndpointGenerator
from gen_epix.seqdb.api.base import EXCLUDED_PERMISSIONS
from gen_epix.seqdb.domain import command, enum, model


class UserInvitationRequestBody(PydanticBaseModel):
    email: str
    roles: set[enum.Role] = Field(description="The roles of the user", min_length=1)
    organization_id: UUID


def create_auth_endpoints(
    router: APIRouter | FastAPI,
    app: App,
    registered_user_dependency: Callable | None = None,
    new_user_dependency: Callable | None = None,
    idp_user_dependency: Callable | None = None,
    handle_exception: Callable | None = None,
    **kwargs: dict,
) -> None:
    assert handle_exception

    # Specific endpoints - Auth
    @router.get(
        "/identity_providers",
        operation_id="identity_providers__get_all",
        name="IdentityProvider",
    )
    async def identity_providers__get_all() -> list[model.IdentityProvider]:
        try:
            cmd = command.GetIdentityProvidersCommand(user=None)
            retval = app.handle(cmd)
        except Exception as exception:
            handle_exception("3ddf8ebb", None, exception)
        return retval

    @router.post(
        "/user_invitations",
        operation_id="user_invitations__post_one",
        name="UserInvitations",
    )
    async def user_invitations__post_one(
        user: registered_user_dependency, user_invitation: UserInvitationRequestBody  # type: ignore
    ) -> model.UserInvitation:
        try:
            retval = app.handle(
                command.InviteUserCommand(
                    user=user,
                    email=user_invitation.email,
                    roles=user_invitation.roles,
                    organization_id=user_invitation.organization_id,
                )
            )
        except Exception as exception:
            handle_exception("e088de91", None, exception)
        return retval

    @router.post(
        "/user_registrations/{token}",
        operation_id="user_registrations__post_one",
        name="RegisterInvitedUser",
    )
    async def user_registrations__post_one(
        user: new_user_dependency, token: str  # type: ignore
    ) -> model.User:
        try:
            cmd = command.RegisterInvitedUserCommand(
                user=user,
                token=token,
            )
            retval = app.handle(cmd)
        except Exception as exception:
            handle_exception("fc1fc53c", None, exception)
        return retval

    # CRUD
    crud_endpoint_sets = CrudEndpointGenerator.create_crud_endpoint_set_for_domain(
        app,
        service_type=enum.ServiceType.AUTH,
        user_dependency=registered_user_dependency,
        excluded_permissions=EXCLUDED_PERMISSIONS,
    )
    CrudEndpointGenerator.generate_endpoints(
        router, crud_endpoint_sets, handle_exception
    )



# pylint: disable=too-few-public-methods


from typing import Any, Type
from uuid import UUID

import sqlalchemy as sa
from sqlalchemy.orm import Mapped

from gen_epix.casedb.domain import enum, model
from gen_epix.casedb.repositories.sa_model.base import RowMetadataMixin
from gen_epix.casedb.repositories.sa_model.util import (
    create_mapped_column,
    create_table_args,
)

Base: Type = sa.orm.declarative_base(name=enum.ServiceType.SUBJECT.value)


class Subject(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Subject)

    data_collection_id: Mapped[UUID] = create_mapped_column(
        model.Subject, "data_collection_id"
    )
    external_ids: Mapped[dict[UUID, str] | None] = create_mapped_column(
        model.Subject, "external_ids"
    )
    content: Mapped[dict[str, Any]] = create_mapped_column(model.Subject, "content")


class SubjectIdentifier(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.SubjectIdentifier)

    subject_id: Mapped[UUID] = create_mapped_column(
        model.SubjectIdentifier, "subject_id"
    )
    identifier_issuer_id: Mapped[UUID] = create_mapped_column(
        model.SubjectIdentifier, "identifier_issuer_id"
    )
    identifier: Mapped[str] = create_mapped_column(
        model.SubjectIdentifier, "identifier"
    )



from typing import Any
from uuid import UUID

import sqlalchemy as sa
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql import expression
from sqlalchemy.sql.compiler import SQLCompiler
from sqlalchemy.types import DateTime, TypeEngine
from sqlalchemy_utils.types.uuid import UUIDType
from sqlmodel._compat import get_sa_type_from_type_annotation
from sqlmodel.main import get_sqlalchemy_type


class ServerUtcTimestamp(expression.FunctionElement):
    type = sa.TIMESTAMP()
    inherit_cache = True


@compiles(ServerUtcTimestamp, "postgresql")
def postgresql_utc_timestamp(
    _element: ServerUtcTimestamp, _compiler: SQLCompiler, **_kw: dict
) -> str:
    return "TIMEZONE('utc', CURRENT_TIMESTAMP)"


@compiles(ServerUtcTimestamp, "mssql")
def mssql_utc_timestamp(
    _element: ServerUtcTimestamp, _compiler: SQLCompiler, **_kw: dict
) -> str:
    return "GETUTCDATE()"


@compiles(ServerUtcTimestamp, "sqlite")
def sqlite_utc_timestamp(
    _element: ServerUtcTimestamp, _compiler: SQLCompiler, **_kw: dict
) -> str:
    return "STRFTIME('%Y-%m-%d %H:%M:%f', 'NOW')"


class ServerUtcCurrentTime(expression.FunctionElement):
    type = DateTime()
    inherit_cache = True


@compiles(ServerUtcCurrentTime, "postgresql")
def postgresql_utc_current_time(
    _element: ServerUtcCurrentTime, _compiler: SQLCompiler, **_kw: dict
) -> str:
    return "TIMEZONE('utc', CURRENT_TIMESTAMP)"


@compiles(ServerUtcCurrentTime, "mssql")
def mssql_utc_current_time(
    _element: ServerUtcCurrentTime, _compiler: SQLCompiler, **_kw: dict
) -> str:
    return "GETUTCDATE()"


@compiles(ServerUtcCurrentTime, "sqlite")
def sqlite_utc_current_time(
    _element: ServerUtcCurrentTime, _compiler: SQLCompiler, **_kw: dict
) -> str:
    return "CURRENT_TIMESTAMP"


def get_pydantic_field_sa_type(fieldinfo: Any) -> TypeEngine:
    """
    Return a suitable SQLAlchemy type for a Pydantic field.
    """
    type_ = get_sa_type_from_type_annotation(fieldinfo.annotation)

    if issubclass(type_, (dict, list, set, tuple)):
        return sa.JSON()

    if issubclass(type_, UUID):
        # Use UUIDType from sqlalchemy_utils to store UUID correctly in all backends
        return UUIDType()

    return get_sqlalchemy_type(fieldinfo)



import abc
import uuid

from gen_epix.fastapp import BaseService
from gen_epix.seqdb.domain import command, model
from gen_epix.seqdb.domain.enum import ServiceType
from gen_epix.seqdb.domain.repository.organization import BaseOrganizationRepository


class BaseOrganizationService(BaseService):
    SERVICE_TYPE = ServiceType.ORGANIZATION

    UPDATE_USER_COMMANDS = {
        command.InviteUserCommand,
        command.UpdateUserCommand,
    }

    # Property overridden to provide narrower return value to support linter
    @property  # type: ignore
    def repository(self) -> BaseOrganizationRepository:  # type: ignore
        return super().repository  # type: ignore

    def register_handlers(self) -> None:
        f = self.app.register_handler
        for command_class in self.app.domain.get_crud_commands_for_service_type(
            self.service_type
        ):
            f(command_class, self.crud)
        f(command.InviteUserCommand, self.invite_user)
        f(command.RegisterInvitedUserCommand, self.register_invited_user)
        f(command.RetrieveCompleteUserCommand, self.retrieve_complete_user)
        f(command.UpdateUserCommand, self.update_user)

    @abc.abstractmethod
    def register_policies(self) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def retrieve_user_by_key(self, user_key: str) -> model.User:
        raise NotImplementedError()

    @abc.abstractmethod
    def invite_user(
        self,
        cmd: command.InviteUserCommand,
    ) -> model.UserInvitation:
        raise NotImplementedError

    @abc.abstractmethod
    def register_invited_user(
        self, cmd: command.RegisterInvitedUserCommand
    ) -> model.User:
        raise NotImplementedError

    def generate_user_invitation_token(self, **kwargs: dict) -> str:
        return str(uuid.uuid4())

    @abc.abstractmethod
    def retrieve_complete_user(
        self, cmd: command.RetrieveCompleteUserCommand
    ) -> model.organization.CompleteUser:
        raise NotImplementedError

    @abc.abstractmethod
    def update_user(
        self,
        cmd: command.UpdateUserCommand,
    ) -> model.User:
        raise NotImplementedError



# pylint: disable=useless-import-alias
from typing import Type

from gen_epix.casedb.domain import enum
from gen_epix.casedb.domain.service.abac import BaseAbacService as BaseAbacService
from gen_epix.casedb.domain.service.case import BaseCaseService as BaseCaseService
from gen_epix.casedb.domain.service.geo import BaseGeoService as BaseGeoService
from gen_epix.casedb.domain.service.ontology import (
    BaseOntologyService as BaseOntologyService,
)
from gen_epix.casedb.domain.service.organization import (
    BaseOrganizationService as BaseOrganizationService,
)
from gen_epix.casedb.domain.service.rbac import BaseRbacService as BaseRbacService
from gen_epix.casedb.domain.service.seqdb import BaseSeqdbService as BaseSeqdbService
from gen_epix.casedb.domain.service.subject import (
    BaseSubjectService as BaseSubjectService,
)
from gen_epix.casedb.domain.service.system import BaseSystemService as BaseSystemService
from gen_epix.fastapp import BaseService
from gen_epix.fastapp.services.auth import BaseAuthService as BaseAuthService

ORDERED_SERVICE_TYPES: list[enum.ServiceType] = [
    enum.ServiceType.GEO,
    enum.ServiceType.ONTOLOGY,
    enum.ServiceType.ORGANIZATION,
    enum.ServiceType.AUTH,
    enum.ServiceType.SEQDB,
    enum.ServiceType.SUBJECT,
    enum.ServiceType.CASE,
    enum.ServiceType.ABAC,
    enum.ServiceType.SYSTEM,
    enum.ServiceType.RBAC,
]

BASE_SERVICE_CLASS_MAP: dict[enum.ServiceType, Type[BaseService]] = {
    enum.ServiceType.GEO: BaseGeoService,
    enum.ServiceType.ONTOLOGY: BaseOntologyService,
    enum.ServiceType.ORGANIZATION: BaseOrganizationService,
    enum.ServiceType.AUTH: BaseAuthService,
    enum.ServiceType.SEQDB: BaseSeqdbService,
    enum.ServiceType.SUBJECT: BaseSubjectService,
    enum.ServiceType.CASE: BaseCaseService,
    enum.ServiceType.ABAC: BaseAbacService,
    enum.ServiceType.SYSTEM: BaseSystemService,
    enum.ServiceType.RBAC: BaseRbacService,
}



from sqlalchemy import Engine

from gen_epix.fastapp.repositories import SARepository
from gen_epix.omopdb.domain.repository.omop import BaseOmopRepository
from gen_epix.omopdb.repositories.sa_model.base import (
    DB_METADATA_FIELDS,
    GENERATE_SERVICE_METADATA,
    SERVICE_METADATA_FIELDS,
)


class OmopSARepository(SARepository, BaseOmopRepository):
    def __init__(self, engine: Engine, **kwargs: dict):
        entities = kwargs.pop("entities", BaseOmopRepository.ENTITIES)
        super().__init__(
            engine,
            entities=entities,
            service_metadata_fields=SERVICE_METADATA_FIELDS,
            db_metadata_fields=DB_METADATA_FIELDS,
            generate_service_metadata=GENERATE_SERVICE_METADATA,
            **kwargs,
        )



# pylint: disable=unused-import-alias
import traceback
from typing import Any, Callable, Type

from uvicorn.logging import logging

from gen_epix.fastapp import App, BaseService
from gen_epix.fastapp.repositories import DictRepository, SARepository
from gen_epix.fastapp.repository import BaseRepository
from gen_epix.fastapp.services.auth import AuthService
from gen_epix.fastapp.services.auth import (
    OIDCClient as OIDCClient,  # pylint: disable=unused-import,useless-import-alias
)
from gen_epix.seqdb.domain import DOMAIN, enum
from gen_epix.seqdb.domain.command.role import RoleGenerator
from gen_epix.seqdb.domain.service import ORDERED_SERVICE_TYPES
from gen_epix.seqdb.repositories import (
    OrganizationDictRepository,
    OrganizationSARepository,
    SeqDictRepository,
    SeqSARepository,
    SystemDictRepository,
    SystemSARepository,
)
from gen_epix.seqdb.services import (
    AbacService,
    OrganizationService,
    RbacService,
    SeqService,
    SystemService,
    UserManager,
)
from util.cfg import AppCfg
from util.env import BaseAppEnv


class AppEnv(BaseAppEnv):
    def __init__(self, app_cfg: AppCfg, log_setup: bool = True, **kwargs: dict):
        self._cfg = app_cfg.cfg
        data = self.compose_application(app_cfg, log_setup=log_setup, **kwargs)
        self._app: App = data["app"]
        self._services: dict[enum.ServiceType, BaseService] = data["services"]
        self._repositories: dict[enum.RepositoryType, BaseRepository] = data[
            "repositories"
        ]
        self._registered_user_dependency: Callable = data["registered_user_dependency"]
        self._new_user_dependency: Callable = data["new_user_dependency"]
        self._idp_user_dependency: Callable = data["idp_user_dependency"]

    @staticmethod
    def compose_application(
        app_cfg: AppCfg, log_setup: bool = True, **kwargs: dict
    ) -> dict:
        # Get logger for setup
        try:
            cfg = app_cfg.cfg
            setup_logger = app_cfg.setup_logger
            app_logger = app_cfg.app_logger
            service_logger = app_cfg.service_logger
            if log_setup:
                setup_logger.debug(
                    App.create_static_log_message(
                        "e8665136", "Starting composing application"
                    )
                )

                setup_logger.debug(
                    App.create_static_log_message(
                        "fb612692", "Initialising services and repositories"
                    )
                )

            # Initialize app
            app = App(
                name="main",
                domain=kwargs.get("domain", DOMAIN),
                logger=app_logger if log_setup else None,
                id_factory=cfg.service.defaults.id_factory,
            )
            if not log_setup:
                app.logger = logging.getLogger(f"seqdb.app")

            # Compose data to initialize repositories and services
            service_data: dict[enum.ServiceType, dict[str, Any]] = {
                enum.ServiceType.ORGANIZATION: {
                    "service_class": OrganizationService,
                    "repository_class": {
                        enum.RepositoryType.DICT: OrganizationDictRepository,
                        enum.RepositoryType.SA_SQL: OrganizationSARepository,
                    },
                },
                enum.ServiceType.AUTH: {
                    "service_class": AuthService,
                    "kwargs": {
                        "idps_cfg": cfg.IDPS_CONFIG,
                    },
                },
                enum.ServiceType.RBAC: {
                    "service_class": RbacService,
                },
                enum.ServiceType.SEQ: {
                    "service_class": SeqService,
                    "repository_class": {
                        enum.RepositoryType.DICT: SeqDictRepository,
                        enum.RepositoryType.SA_SQL: SeqSARepository,
                    },
                },
                enum.ServiceType.SYSTEM: {
                    "service_class": SystemService,
                    "repository_class": {
                        enum.RepositoryType.DICT: SystemDictRepository,
                        enum.RepositoryType.SA_SQL: SystemSARepository,
                    },
                },
                enum.ServiceType.ABAC: {
                    "service_class": AbacService,
                },
            }
            for service_type in service_data:
                if "repository_class" in service_data[service_type]:
                    service_data[service_type]["repository_class"][
                        enum.RepositoryType.SA_SQLITE
                    ] = service_data[service_type]["repository_class"][
                        enum.RepositoryType.SA_SQL
                    ]

            # Initialise repositories and services
            services: dict[enum.ServiceType, BaseService] = {}
            repositories: dict[enum.ServiceType, BaseRepository] = {}
            for service_type in ORDERED_SERVICE_TYPES:
                data = service_data[service_type]
                props = {
                    x: y
                    for x, y in cfg.service[service_type.value.lower()].items()
                    if x not in {"id_factory", "timestamp_factory"}
                }
                id_factory = cfg.service[service_type.value.lower()]["id_factory"]
                timestamp_factory = cfg.service[service_type.value.lower()][
                    "timestamp_factory"
                ]
                additional_service_kwargs: dict = data.get("kwargs", {})  # type: ignore

                # Create repository if necessary
                if "repository_class" in data:
                    entities = app.domain.get_dag_sorted_entities(
                        service_type=service_type
                    )
                    repository_type = cfg.secret["db"]["repository_type"]
                    repository_cfg = cfg.secret["repository"][
                        repository_type.value.lower()
                    ][service_type.value.lower()]
                    if log_setup:
                        setup_logger.debug(
                            app.create_log_message(
                                "db89f0a5",
                                f"Setting up {service_type.value} service with {repository_type.value} repository",
                            )
                        )
                    repository_class = data["repository_class"][repository_type]
                    if repository_type == enum.RepositoryType.DICT:
                        curr_repository = DictRepository.from_pkl(
                            repository_class,  # type: ignore
                            entities,
                            repository_cfg["file"],
                            timestamp_factory=timestamp_factory,
                        )
                    elif repository_type == enum.RepositoryType.SA_SQLITE:
                        assert issubclass(repository_class, SARepository)
                        curr_repository = repository_class.create_sa_repository(
                            entities,
                            "sqlite:///" + repository_cfg["file"],
                            name=service_type.value,
                            timestamp_factory=timestamp_factory,
                        )
                    elif repository_type == enum.RepositoryType.SA_SQL:
                        assert issubclass(repository_class, SARepository)
                        curr_repository = repository_class.create_sa_repository(
                            entities,
                            repository_cfg["connection_string"],
                            name=service_type.value,
                            timestamp_factory=timestamp_factory,
                        )
                    else:
                        raise NotImplementedError()
                else:
                    curr_repository = None
                # Create service, injecting app, repository, logger and props
                service_class: Type[BaseService] = data["service_class"]
                curr_service: BaseService = service_class(
                    app,
                    service_type=service_type,
                    repository=curr_repository,
                    logger=setup_logger if log_setup else None,
                    props=props,
                    name=service_type.value,
                    id_factory=id_factory,
                    **additional_service_kwargs,
                )
                if not log_setup:
                    curr_service.logger = service_logger
                # Add to overview of services and repositories
                repositories[service_type] = curr_repository
                services[service_type] = curr_service

            # Register roles
            service = services[enum.ServiceType.RBAC]
            assert isinstance(service, RbacService)
            service.register_roles(RoleGenerator.ROLE_PERMISSIONS)

            # Create and set user generator, which can create new users under different scenarios
            # such as from claims, from invitation, and when matching root secret
            app.user_manager = UserManager(
                services[enum.ServiceType.ORGANIZATION],  # type: ignore
                services[enum.ServiceType.RBAC],  # type: ignore
                cfg.secret.root,
                automatic_new_user_cfg=cfg.secret.automatic_new_user,  # set to None if no automatic new user
            )

            # Get current user and new user dependencies for injecting authentication in endpoints
            registered_user_dependency, new_user_dependency, idp_user_dependency = services[enum.ServiceType.AUTH].create_user_dependencies()  # type: ignore

            # Register security policies with app
            if log_setup:
                setup_logger.debug(
                    app.create_log_message("f329be4d", "Registering security policies")
                )
            services[enum.ServiceType.RBAC].register_policies()  # type: ignore
            services[enum.ServiceType.ABAC].register_policies()  # type: ignore
            services[enum.ServiceType.SYSTEM].register_policies()  # type: ignore

            # Finalise process
            if log_setup:
                setup_logger.debug(
                    app.create_log_message("da172304", "Finished composing application")
                )

        except Exception as e:

            # Print error for deployment log, in regular log is not shown there
            print(traceback.print_exc())
            if log_setup:
                setup_logger.error(
                    App.create_static_log_message(
                        "db960800",
                        f"Error setting up application: {traceback.print_exc()}",
                    )
                )
            raise e

        return {
            "app": app,
            "services": services,
            "repositories": repositories,
            "registered_user_dependency": registered_user_dependency,
            "new_user_dependency": new_user_dependency,
            "idp_user_dependency": idp_user_dependency,
        }



from gen_epix.fastapp import PermissionTypeSet
from gen_epix.omopdb.domain import model

EXCLUDED_PERMISSIONS: dict = {
    model.User: PermissionTypeSet.CU,
    model.UserInvitation: PermissionTypeSet.CU,
}



# Perhaps there's a better location for this?
# Also best practice might be rename all enum.py files enums.py to ensure avoiding name collisions with built-in enum module.

from enum import Enum


class AnonStrictness(Enum):
    IGNORE = "ignore"
    WARN = "warn"
    STRICT = "strict"


class AnonMethod(Enum):
    MAKE_NULL = "make_null"
    SHIFT = "shift"
    RANDOM = "random"
    CATEGORICAL = "categorical"
    MODEL_ANONYIMIZATION = "model_anonymization"  # for future use



# pylint: disable=too-few-public-methods


import json
from datetime import datetime
from typing import ClassVar
from uuid import UUID

from pydantic import Field, field_serializer, field_validator

from gen_epix.fastapp import Permission
from gen_epix.fastapp import User as ServiceUser
from gen_epix.fastapp.domain import Entity, create_keys, create_links
from gen_epix.seqdb.domain import DOMAIN, enum
from gen_epix.seqdb.domain.model.base import Model

_SERVICE_TYPE = enum.ServiceType.ORGANIZATION
_ENTITY_KWARGS = {
    "service_type": _SERVICE_TYPE,
    "schema_name": _SERVICE_TYPE.value.lower(),
}


class UserNameEmail(ServiceUser, Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="user_name_emails",
        persistable=False,
        **_ENTITY_KWARGS,
    )
    id: UUID | None = Field(default=None, description="The ID of the user")
    name: str | None = Field(
        default=None, description="The full name of the user", max_length=255
    )
    email: str = Field(description="The email of the user", max_length=320)


class User(ServiceUser, Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="users",
        table_name="user",
        persistable=True,
        keys=create_keys({1: "email"}),
        **_ENTITY_KWARGS,
    )
    id: UUID | None = Field(default=None, description="The ID of the user")
    email: str = Field(description="The email of the user, UNIQUE", max_length=320)
    name: str | None = Field(
        default=None, description="The full name of the user", max_length=255
    )

    is_active: bool = Field(
        default=True,
        description="Whether the user is active or not. An inactive user cannot perform any actions that require authorization.",
    )
    roles: set[enum.Role] = Field(description="The roles of the user", min_length=1)
    data_collection_ids: set[UUID] = Field(
        description="The data collections the user has access to"
    )

    def get_key(self) -> str:
        return self.email

    @field_validator("roles", mode="before")
    @classmethod
    def _validate_roles(cls, value: set[enum.Role] | list[str] | str) -> set[enum.Role]:
        """
        Validate and convert roles representation to a set[Role]. When given as a
        string, it is assumed to be a JSON list of Role values.
        """
        if isinstance(value, str):
            return {enum.Role[x] for x in json.loads(value)}
        if isinstance(value, list):
            return {enum.Role[x] for x in value}
        return value

    @field_validator("data_collection_ids", mode="before")
    @classmethod
    def _validate_data_collection_ids(
        cls, value: set[UUID] | list[str] | str
    ) -> set[UUID]:
        """
        Validate and convert data_collection_ids representation to a set[UUID].
        When given as a string, it is assumed to be a JSON list of UUIDs.
        """
        if isinstance(value, str):
            return {UUID(x) for x in json.loads(value)}
        if isinstance(value, list):
            return {UUID(x) for x in value}
        return value

    @field_serializer("roles")
    def serialize_roles(self, value: set[enum.Role], _info):
        return [x.value for x in value]

    @field_serializer("data_collection_ids")
    def serialize_data_collection_ids(self, value: set[UUID], _info):
        return [str(x) for x in value]


class UserInvitation(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="user_invitations",
        table_name="user_invitation",
        persistable=True,
        keys=create_keys({1: ("email", "expires_at")}),
        links=create_links(
            {
                1: ("invited_by_user_id", User, "invited_by_user"),
            }
        ),
        **_ENTITY_KWARGS,
    )
    email: str = Field(description="The email of the user, UNIQUE", max_length=320)
    token: str = Field(description="The token of the invitation", max_length=255)
    expires_at: datetime = Field(description="The expiry date of the invitation")
    roles: set[enum.Role] = Field(description="The initial roles of the user")
    data_collection_ids: set[UUID] = Field(
        description="The data collections the user has access to"
    )
    invited_by_user_id: UUID = Field(
        description="The ID of the user who invited the user. FOREIGN KEY"
    )
    invited_by_user: User | None = Field(
        default=None, description="The user who invited the user"
    )


class IdentifierIssuer(Model):
    """
    A system or process that issues identifiers.
    The combination (identifier_issuer, issued_identifier) is universally unique.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="identifier_issuers",
        table_name="identifier_issuer",
        persistable=True,
        **_ENTITY_KWARGS,
    )
    name: str = Field(description="The name of the issuer", max_length=255)


class DataCollection(Model):
    """
    Represents a collection of data.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="data_collections",
        table_name="data_collection",
        persistable=True,
        keys=create_keys({1: "name"}),
        **_ENTITY_KWARGS,
    )
    # TODO: Placeholder
    name: str = Field(
        description="The name of a data collection, UNIQUE", max_length=255
    )
    description: str | None = Field(
        default=None, description="The description of the data collection."
    )


class DataCollectionSet(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="data_collection_sets",
        table_name="data_collection_set",
        persistable=True,
        keys=create_keys({1: "name"}),
        **_ENTITY_KWARGS,
    )
    name: str = Field(description="The name of the data collection set", max_length=255)
    description: str | None = Field(
        default=None, description="The description of the data collection set."
    )


class DataCollectionSetMember(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="data_collection_set_members",
        table_name="data_collection_set_member",
        persistable=True,
        keys=create_keys(
            {1: ("data_collection_set_id", "data_collection_id")},
        ),
        links=create_links(
            {
                1: ("data_collection_set_id", DataCollectionSet, "data_collection_set"),
                2: ("data_collection_id", DataCollection, "data_collection"),
            }
        ),
        **_ENTITY_KWARGS,
    )
    data_collection_set_id: UUID = Field(
        description="The ID of the data collection set. FOREIGN KEY"
    )
    data_collection_set: DataCollectionSet | None = Field(
        default=None, description="The data collection set"
    )
    data_collection_id: UUID = Field(
        description="The ID of the data collection. FOREIGN KEY"
    )
    data_collection: DataCollection | None = Field(
        default=None, description="The data collection"
    )


class CompleteUser(User):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="complete_users",
        persistable=False,
        **_ENTITY_KWARGS,
    )
    roles: set[enum.Role] = Field(description="The roles of the user")
    permissions: set[Permission] = Field(
        description="The union of all the permissions of the user"
    )
    data_collection_ids: set[UUID] = Field(
        description="The ids of the data collections the user has access to"
    )
    data_collections: set[DataCollection] = Field(
        description="The data collections the user has access to"
    )


DOMAIN.register_locals(locals(), service_type=_SERVICE_TYPE)



import logging
from typing import Any, Callable

from fastapi import FastAPI, Response
from fastapi.concurrency import asynccontextmanager
from fastapi.responses import RedirectResponse
from slowapi import _rate_limit_exceeded_handler
from slowapi.errors import RateLimitExceeded
from slowapi.middleware import SlowAPIMiddleware

from gen_epix.casedb.api.router import create_routers
from gen_epix.common.api.exc import generate_handle_exception_function
from gen_epix.fastapp import App
from gen_epix.fastapp.api.openapi import create_custom_openapi_function
from gen_epix.fastapp.middleware import (
    HandleAuthExceptionMiddleware,
    UpdateResponseHeaderMiddleware,
    limiter,
)


def create_fast_api(
    cfg: dict,
    app: App,
    registered_user_dependency: Callable | None = None,
    new_user_dependency: Callable | None = None,
    idp_user_dependency: Callable | None = None,
    handle_exception: Callable | None = None,
    setup_logger: logging.Logger | None = None,
    api_logger: logging.Logger | None = None,
    debug: bool = False,
    **kwargs: dict,
) -> FastAPI:

    app_id = kwargs.pop("app_id", app.generate_id())

    # Set up lifespan
    @asynccontextmanager
    async def lifespan(fast_api: FastAPI) -> Any:
        if setup_logger:
            setup_logger.info(
                app.create_log_message(
                    "a49dedfc",
                    {"status": "STARTING_APP", "app_id": str(app_id)},  # type: ignore[arg-type]
                )
            )
        yield
        if setup_logger:
            setup_logger.info(
                app.create_log_message(
                    "dcabb0ac",
                    {"status": "STOPPING_APP", "app_id": str(app_id)},  # type: ignore[arg-type]
                )
            )

    # Initialize fast_api
    fast_api = FastAPI(
        separate_input_output_schemas=False,
        swagger_ui_init_oauth={"usePkceWithAuthorizationCodeGrant": True},
        openapi_tags=kwargs.get(  # type: ignore[arg-type]
            "openapi_tags",
            [
                {
                    "name": "core",
                    "description": "Core functionality",
                }
            ],
        ),
        lifespan=lifespan,
    )

    # Add middleware
    # Rate limiting
    if not debug:
        fast_api.state.limiter = limiter.limiter
        fast_api.add_exception_handler(RateLimitExceeded, _rate_limit_exceeded_handler)  # type: ignore[arg-type]
        # The SlowAPIMiddleware is added to fast_api globally to limit the number of requests
        # The limiter can be applied to specific routes by adding the decorator @limiter.limit
        fast_api.add_middleware(SlowAPIMiddleware)

    # Response header handling
    if not debug:
        fast_api.add_middleware(
            UpdateResponseHeaderMiddleware,
            general_headers=cfg.api.http_header.general,
            exception_headers=[
                ({"/docs/oauth2-redirect"}, cfg.api.http_header.auth),
                ({"/docs", "/redoc"}, cfg.api.http_header.openapi),
            ],
        )
    # Handling of authentication exceptions
    if not debug:
        fast_api.add_middleware(
            HandleAuthExceptionMiddleware,
            fast_app=app,
            logger=api_logger,
        )

    # Add routers
    handle_exception = generate_handle_exception_function(app=app, logger=api_logger)
    routers = create_routers(
        app=app,
        registered_user_dependency=registered_user_dependency,
        new_user_dependency=new_user_dependency,
        idp_user_dependency=idp_user_dependency,
        handle_exception=handle_exception,
    )
    for router in routers:
        fast_api.include_router(router, prefix=cfg.api.route.v1)

    # Redirect root to default route
    @fast_api.get("/")
    async def redirect() -> Response:
        response = RedirectResponse(url=cfg.api.default_route)
        return response

    # Update OpenAPI schema generator function
    if kwargs.pop("update_openapi_schema", False):
        update_openapi_kwargs = kwargs.pop("update_openapi_kwargs", {})
        get_open_api_kwargs = update_openapi_kwargs.pop("get_openapi_kwargs", {})
        get_open_api_kwargs.update({"routes": fast_api.routes})
        custom_openapi_function = create_custom_openapi_function(
            get_open_api_kwargs,
            fix_schema=update_openapi_kwargs.get("fix_schema", False),
            auth_service=update_openapi_kwargs.get("auth_service"),
        )
        setattr(fast_api, "openapi", custom_openapi_function)

    return fast_api



from gen_epix.fastapp.services.auth import IdentityProvider as IdentityProvider
from gen_epix.fastapp.services.auth import IDPUser as IDPUser

# pylint: disable=useless-import-alias
from gen_epix.seqdb.domain.model.base import Model as Model
from gen_epix.seqdb.domain.model.organization import *
from gen_epix.seqdb.domain.model.seq import *
from gen_epix.seqdb.domain.model.system import *



from typing import Any, Type

import sqlalchemy as sa
from sqlalchemy.orm import MappedColumn, mapped_column

from gen_epix.fastapp import Entity
from gen_epix.fastapp.repositories import get_pydantic_field_sa_type
from gen_epix.omopdb.domain.model import Model
from gen_epix.omopdb.domain.model.base import Model


def create_table_args(
    model_class: Type[Model],
    field_name_map: dict[str, str] | None = None,
    **kwargs: dict,
) -> tuple:
    entity: Entity = model_class.ENTITY  # type: ignore[attr-defined]
    uq_constraints = []
    for field_names in entity.get_keys_field_names():
        sa_field_names = (
            [field_name_map.get(x, x) for x in field_names]
            if field_name_map
            else field_names
        )
        sa_field_name_str = "_".join(sa_field_names)
        uq_constraints.append(
            sa.UniqueConstraint(
                *sa_field_names,
                name=f"uq_{entity.table_name}_{sa_field_name_str}",
                **kwargs,
            )
        )
    if entity.schema_name:
        return entity.table_name, tuple(
            [*uq_constraints, {"schema": entity.schema_name}]
        )
    return entity.table_name, tuple([*uq_constraints])


def create_mapped_column(
    model_class: Type[Model],
    field_name: str,
    field_name_map: dict[str, str] | None = None,
    sql_column_name: str | None = None,
    **kwargs: dict,
) -> MappedColumn[Any]:
    """
    Stub of a docstring reflecting best practice: what do the args do? ;-)
    sql_column_name: optional, if provided passed to sqlalchemy's mapped_column to indicate the column
    in the database that corresponds to this field; not needed if names match
    """
    entity: Entity = model_class.ENTITY  # type: ignore[attr-defined]
    fieldinfo = model_class.model_fields[field_name]
    sa_type = get_pydantic_field_sa_type(fieldinfo)
    nullable = kwargs.get("nullable", fieldinfo.is_required() is False)
    doc = kwargs.pop("doc", fieldinfo.description)
    sql_column_name = kwargs.pop("sql_column_name", None)
    link_entity = entity.get_link_entity(field_name)
    if link_entity and link_entity.service_type != entity.service_type:
        # Create foreign keys only within the same service
        link_entity = None
    ondelete = kwargs.pop("ondelete", None)
    onupdate = kwargs.pop("onupdate", None)
    sa_field_name = (
        field_name_map[model_class][field_name] if field_name_map else field_name
    )
    fk_name = kwargs.pop("fk_name", f"fk_{entity.table_name}_{sa_field_name}")
    # Build up the args/kwargs we pass to sqlalchemy's mapped_column
    # This is necessary because mapped_column takes certain information as positional-only arguments
    # but not all are actually required (look at the sqlalchemy code!), and we only
    # have cases where we need to pass and cases where we don't need to pass!
    args = []
    extra_kwargs = {}
    if sql_column_name:
        args.append(sql_column_name)
        extra_kwargs["key"] = field_name
    args.append(sa_type)
    if link_entity:
        link_model_class = link_entity.model_class
        link_sa_id_field_name = (
            field_name_map[link_model_class][link_entity.id_field_name]
            if field_name_map
            else link_entity.id_field_name
        )
        ref_column_name = (
            f"{link_entity.schema_name}.{link_entity.table_name}.{link_sa_id_field_name}"
            if link_entity.schema_name
            else f"{link_entity.table_name}.{link_sa_id_field_name}"
        )
        fk_arg = sa.ForeignKey(
            ref_column_name, ondelete=ondelete, onupdate=onupdate, name=fk_name
        )
        args.append(fk_arg)
        # return mapped_column(
        #    sa_type,
        #    sa.ForeignKey(
        #        ref_column_name, ondelete=ondelete, onupdate=onupdate, name=fk_name
        #    ),
        #    nullable=nullable,
        #    doc=doc,
        #    **kwargs,
        # )
    else:
        extra_kwargs["primary_key"] = entity.id_field_name == field_name
    passed_kwargs = {"nullable": nullable, "doc": doc, **kwargs, **extra_kwargs}
    # return mapped_column(
    #    sa_type,
    #    nullable=nullable,
    #    primary_key=entity.id_field_name == field_name,
    #    doc=doc,
    #    **kwargs,
    # )
    return mapped_column(*args, **passed_kwargs)



from uuid import UUID

import numpy as np
from sqlalchemy import Engine

from gen_epix.fastapp import BaseUnitOfWork, CrudOperation
from gen_epix.fastapp.repositories import SARepository
from gen_epix.seqdb.domain import model
from gen_epix.seqdb.domain.repository.seq import BaseSeqRepository
from gen_epix.seqdb.repositories.sa_model.base import (
    DB_METADATA_FIELDS,
    GENERATE_SERVICE_METADATA,
    SERVICE_METADATA_FIELDS,
)


class SeqSARepository(SARepository, BaseSeqRepository):
    def __init__(self, engine: Engine, **kwargs: dict):
        entities = kwargs.pop("entities", BaseSeqRepository.ENTITIES)
        super().__init__(
            engine,
            entities=entities,
            service_metadata_fields=SERVICE_METADATA_FIELDS,
            db_metadata_fields=DB_METADATA_FIELDS,
            generate_service_metadata=GENERATE_SERVICE_METADATA,
            **kwargs,
        )

    def get_distance_matrix_by_seq_ids(
        self,
        uow: BaseUnitOfWork,
        seq_distance_protocol_id: UUID,
        seq_ids: list[UUID],
    ) -> np.ndarray:
        raise NotImplementedError("Code to be converted to seqdb architecture")
        self.raise_on_duplicate_ids(seq_ids)
        seqs = self.crud(
            uow,
            None,
            model.SeqDistance,
            None,
            seq_ids,
            CrudOperation.READ_SOME,
        )
        id_to_idx_map = {x.id: i for i, x in enumerate(seqs)}
        n = len(seqs)
        distance_matrix = np.empty((n, n))
        distance_matrix[:] = np.nan
        for i in range(n):
            for id_, distance in seqs[i].distances.items():
                if id_ not in id_to_idx_map:
                    continue
                distance_matrix[id_to_idx_map[id_], i] = distance
            distance_matrix[i, i] = 0
        return distance_matrix



from gen_epix.casedb.domain.repository.case import BaseCaseRepository
from gen_epix.fastapp.repositories import DictRepository


class CaseDictRepository(DictRepository, BaseCaseRepository):
    pass



import abc
from uuid import UUID

from fastapi import Request

from gen_epix.fastapp.services.auth.model import Claims, IdentityProvider


class IDPClient(abc.ABC):

    @property
    @abc.abstractmethod
    def id(self) -> UUID:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_identity_provider(self) -> IdentityProvider:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_claims_from_userinfo(
        self, access_token: str
    ) -> dict[str, str | int | bool | list[str]]:
        raise NotImplementedError()

    @abc.abstractmethod
    async def __call__(self, request: Request) -> Claims | None:
        """
        Returns the claims of the user from the request or None if claims cannot be
        processed by this client.
        """
        raise NotImplementedError()



from pkg_resources import get_distribution

from gen_epix.seqdb.app_setup import create_fast_api
from gen_epix.seqdb.domain import enum
from gen_epix.seqdb.env import AppEnv
from util.cfg import AppCfg

APP_NAME = "SEQDB"

# Data for OpenAPI schema
SCHEMA_KWARGS = {
    "title": "Sequence database",
    "description": "Sequence database API",
    "summary": "Summary goes here",
    "version": get_distribution("gen-epix").version,
    "terms_of_service": "http://example.com/terms/",
    "contact": {
        "name": "RIVM CIb IDS bioinformatics group",
        "url": "https://github.com/RIVM-bioinformatics/gen-epix-api",
        "email": "ids-bioinformatics@rivm.nl",
    },
    "license_info": {
        "name": "License to be confirmed",
        "identifier": "Apache-2.0",
    },
}

# Get configuration data and environment
APP_CFG = AppCfg(APP_NAME, enum.ServiceType, enum.RepositoryType)
APP_ENV = AppEnv(APP_CFG)

# Create fastapi
FAST_API = create_fast_api(
    APP_CFG.cfg,
    app=APP_ENV.app,
    registered_user_dependency=APP_ENV.registered_user_dependency,
    new_user_dependency=APP_ENV.new_user_dependency,
    idp_user_dependency=APP_ENV.idp_user_dependency,
    app_id=APP_ENV.app.generate_id(),
    setup_logger=APP_CFG.setup_logger,
    api_logger=APP_CFG.api_logger,
    debug=APP_CFG.cfg.app.debug,
    update_openapi_schema=True,
    update_openapi_kwargs={
        "get_openapi_kwargs": SCHEMA_KWARGS,
        "fix_schema": True,
        "auth_service": APP_ENV.services[enum.ServiceType.AUTH],
    },
)

# TODO: app variable added for backwards compatibility with startup code that imports "app". Remove once that code is updated as well.
app = FAST_API
